{{ '"""{}"""'.format(obj.help | clean_docstring) }}
{% if obj.has_help_attr %}
{#
    Note mkdocstrings-material wants "Attributes rather than
    "Parameters" or "Properties" when there is no function
    signature, like dataclasses and pydantic.
#}
Attributes
----------
{%- for var_name, var_type, var_doc in obj | class_params %}
{{ "{} : {}".format(var_name, var_type) | text_wrap() }}
{{ "    {}".format(var_doc) | text_wrap() }}
{%- endfor -%}
{%- endif %}
